#include <qpixmap.h>
#include <qstyle.h>
#include <qpainter.h>
#include <qpushbutton.h>
#include <qimage.h>
#include <qstylefactory.h>
#include <qapplication.h>
#include <qtabbar.h>
#include <qcheckbox.h>
#include <qradiobutton.h>
#include <qscrollbar.h>
#include <qmenubar.h>
#include <qmenudata.h>

#include <stdlib.h>

#include <gtk/gtknotebook.h>
#include <gtk/gtkgc.h>

#include "qt_qt_wrapper.h"

class BufferedImage
{
public:
	enum WidgetType
	{
		Button,
		Tab,
		LineEdit,
		CheckBox,
		RadioButton,
		MenuItem,
		MenuBar
	};

	BufferedImage() {};
	BufferedImage(WidgetType t, GtkStateType s, int w, int h, QImage i);
	BufferedImage(WidgetType t, GtkStateType s, int w, int h, QImage i, int extra);

	WidgetType type;
	GtkStateType state_type;
	int width;
	int height;
	QImage image;

	int on; // For Check Boxes
	int hasFocus; // For Line Edits
};

QScrollBar* meepBar;
QValueList<BufferedImage> buffer;

BufferedImage::BufferedImage(WidgetType t, GtkStateType s, int w, int h, QImage i)
{
	type = t;
	state_type = s;
	width = w;
	height = h;
	image = i;
}

BufferedImage::BufferedImage(WidgetType t, GtkStateType s, int w, int h, QImage i, int extra)
{
	type = t;
	state_type = s;
	width = w;
	height = h;
	image = i;
	if (t == CheckBox) on = extra;
	if (t == LineEdit) hasFocus = extra;
}



void createQApp()
{
	int argc;
	char** argv;
	new QApplication(argc, argv);
	meepBar = new QScrollBar(NULL);
}

void destroyQApp()
{
	delete meepBar;
	delete qApp;
}

void drawButton(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int x, int y, int width, int height)
{
	QValueList<BufferedImage>::iterator it;
    for ( it = buffer.begin(); it != buffer.end(); ++it )
    {
		if (((*it).type == BufferedImage::Button) &&
			((*it).state_type == state_type) &&
			((*it).width == width) &&
			((*it).height == height))
		{
			gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, (*it).image.bits(), (*it).image.bytesPerLine());
			return;
		}
	}

	QPixmap* pixmap = new QPixmap(width,height);
	QPainter* painter = new QPainter(pixmap);
	QPushButton* button = new QPushButton(NULL);

	QStyle::SFlags sflags = QStyle::Style_Default;

	switch (state_type)
	{
	case GTK_STATE_ACTIVE: 		sflags |= QStyle::Style_Down;		break;
	case GTK_STATE_PRELIGHT:	sflags |= QStyle::Style_MouseOver;	break;
	case GTK_STATE_SELECTED:	sflags |= QStyle::Style_HasFocus;	break;
	case GTK_STATE_INSENSITIVE:	button->setEnabled(false);			break;
	}

	qApp->style().drawControl(QStyle::CE_PushButton, painter, button, QRect(0,0,width,height), qApp->palette().active(), sflags);

	QImage image = pixmap->convertToImage().swapRGB();
	delete painter;
	delete pixmap;
	delete button;

	buffer.append(BufferedImage(BufferedImage::Button, state_type, width, height, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawTab(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int x, int y, int width, int height)
{
	QValueList<BufferedImage>::iterator it;
    for ( it = buffer.begin(); it != buffer.end(); ++it )
    {
		if (((*it).type == BufferedImage::Tab) &&
			((*it).state_type == state_type) &&
			((*it).width == width) &&
			((*it).height == height))
		{
			gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, (*it).image.bits(), (*it).image.bytesPerLine());
			return;
		}
	}

	QPixmap* pixmap = new QPixmap(width,height);
	QPainter* painter = new QPainter(pixmap);
	QWidget* tabbarparent = new QWidget(); // Included to stop some styles from complaining
	QTabBar* tabbar = new QTabBar(tabbarparent);

	QStyle::SFlags sflags = QStyle::Style_Default;

	if (state_type != GTK_STATE_ACTIVE)
		sflags = QStyle::Style_Selected;
	else
		sflags = QStyle::Style_Enabled;

	painter->fillRect(0,0,width,height,qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_TabBarTab, painter, tabbar, QRect(0,0,width,height), qApp->palette().active(), sflags);

	QImage image = pixmap->convertToImage().swapRGB();
	delete painter;
	delete pixmap;
	delete tabbar;
	delete tabbarparent;

	buffer.append(BufferedImage(BufferedImage::Tab, state_type, width, height, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawVLine(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int x, int y1, int y2)
{
	int width = 2;
	int height = style->xthickness;

	QPixmap* pixmap = new QPixmap(width, height);
	QPainter* painter = new QPainter(pixmap);

	painter->fillRect(2,0,width-2,height,qApp->palette().active().background());
	painter->setPen( qApp->palette().active().mid() );
	painter->drawLine( 0, 0, 0, height );
	painter->setPen( qApp->palette().active().light() );
	painter->drawLine( 1, 0, 1, height );

	QImage image = pixmap->convertToImage().swapRGB();
	delete painter;
	delete pixmap;

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y1, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawHLine(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int y, int x1, int x2)
{
	int width = abs(x2-x1);
	int height = style->ythickness;

	QPixmap* pixmap = new QPixmap(width, height);
	QPainter* painter = new QPainter(pixmap);

	painter->fillRect(0,2,width,height-2,qApp->palette().active().background());
	painter->setPen( qApp->palette().active().mid() );
	painter->drawLine( 0, 0, width, 0);
	painter->setPen( qApp->palette().active().light() );
	painter->drawLine( 0, 1, width, 1);

	QImage image = pixmap->convertToImage().swapRGB();
	delete painter;
	delete pixmap;

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x1, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawLineEdit(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int hasFocus, int x, int y, int width, int height)
{
	QValueList<BufferedImage>::iterator it;
    for ( it = buffer.begin(); it != buffer.end(); ++it )
    {
		if (((*it).type == BufferedImage::LineEdit) &&
			((*it).state_type == state_type) &&
			((*it).width == width) &&
			((*it).height == height) &&
			((*it).hasFocus == hasFocus))
		{
			gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, (*it).image.bits(), (*it).image.bytesPerLine());
			return;
		}
	}

	QPixmap* pixmap = new QPixmap(width, height);
	QPainter* painter = new QPainter(pixmap);

	QStyle::SFlags sflags = QStyle::Style_Default;
	if (hasFocus)
		sflags = QStyle::Style_HasFocus;

	painter->fillRect(0,0,width,height,qApp->palette().active().background());
	qApp->style().drawPrimitive(QStyle::PE_PanelLineEdit, painter, QRect(0,0,width,height), qApp->palette().active(), sflags);

	QImage image = pixmap->convertToImage().swapRGB();
	delete painter;
	delete pixmap;

	buffer.append(BufferedImage(BufferedImage::LineEdit, state_type, width, height, image, hasFocus));

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawCheckBox(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int on, int x, int y, int width, int height)
{
	QValueList<BufferedImage>::iterator it;
    for ( it = buffer.begin(); it != buffer.end(); ++it )
    {
		if (((*it).type == BufferedImage::CheckBox) &&
			((*it).state_type == state_type) &&
			((*it).width == width) &&
			((*it).height == height) &&
			((*it).on == on))
		{
			gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, (*it).image.bits(), (*it).image.bytesPerLine());
			return;
		}
	}

	QPixmap* pixmap = new QPixmap(width,height);
	QPainter* painter = new QPainter(pixmap);
	QCheckBox* checkbox = new QCheckBox(NULL);

	QStyle::SFlags sflags = QStyle::Style_Default;

	switch (state_type)
	{
	case GTK_STATE_ACTIVE: 		sflags |= QStyle::Style_Down;		break;
	case GTK_STATE_PRELIGHT:	sflags |= QStyle::Style_MouseOver;	break;
	case GTK_STATE_SELECTED:	sflags |= QStyle::Style_HasFocus;	break;
	case GTK_STATE_INSENSITIVE:	checkbox->setEnabled(false);		break;
	}

	if (on)
		sflags |= QStyle::Style_On;
	else
		sflags |= QStyle::Style_Off;

	painter->fillRect(0,0,width,height,qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_CheckBox, painter, checkbox, QRect(0,0,width,height), qApp->palette().active(), sflags);

	QImage image = pixmap->convertToImage().swapRGB();
	delete painter;
	delete pixmap;
	delete checkbox;

	buffer.append(BufferedImage(BufferedImage::CheckBox, state_type, width, height, image, on));

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawRadioButton(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int on, int x, int y, int width, int height)
{
	QValueList<BufferedImage>::iterator it;
    for ( it = buffer.begin(); it != buffer.end(); ++it )
    {
		if (((*it).type == BufferedImage::RadioButton) &&
			((*it).state_type == state_type) &&
			((*it).width == width) &&
			((*it).height == height))
		{
			gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, (*it).image.bits(), (*it).image.bytesPerLine());
			return;
		}
	}

	QPixmap* pixmap = new QPixmap(width,height);
	QPainter* painter = new QPainter(pixmap);
	QRadioButton* radio = new QRadioButton(NULL);

	QStyle::SFlags sflags = QStyle::Style_Default;

	switch (state_type)
	{
	case GTK_STATE_ACTIVE: 		sflags |= QStyle::Style_Down;		break;
	case GTK_STATE_PRELIGHT:	sflags |= QStyle::Style_MouseOver;	break;
	case GTK_STATE_SELECTED:	sflags |= QStyle::Style_HasFocus;	break;
	case GTK_STATE_INSENSITIVE:	radio->setEnabled(false);			break;
	}

	if (on)
		sflags |= QStyle::Style_On;
	else
		sflags |= QStyle::Style_Off;

	painter->fillRect(0,0,width,height,qApp->palette().active().background());
	qApp->style().drawControl(QStyle::CE_RadioButton, painter, radio, QRect(0,0,width,height), qApp->palette().active(), sflags);

	QImage image = pixmap->convertToImage().swapRGB();
	delete painter;
	delete pixmap;
	delete radio;

	buffer.append(BufferedImage(BufferedImage::RadioButton, state_type, width, height, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawScrollBar(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int orientation, GtkAdjustment* adj, int x, int y, int width, int height)
{
	//QScrollBar* sbar = new QScrollBar(0,100,1,10,50,orientation ? Qt::Vertical : Qt::Horizontal, NULL);

	meepBar->setMinValue((int)adj->lower);
	meepBar->setMaxValue((int)adj->upper - adj->page_size);
	meepBar->setValue((int)adj->value);
	meepBar->setPageStep((int)adj->page_size);
	meepBar->setOrientation(orientation ? Qt::Vertical : Qt::Horizontal);
	meepBar->resize(width,height);

	QImage image = QPixmap::grabWidget(meepBar).convertToImage().swapRGB();

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawMenuBarItem(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int x, int y, int width, int height)
{
	QValueList<BufferedImage>::iterator it;
    for ( it = buffer.begin(); it != buffer.end(); ++it )
    {
		if (((*it).type == BufferedImage::MenuItem) &&
			((*it).state_type == state_type) &&
			((*it).width == width) &&
			((*it).height == height))
		{
			gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, (*it).image.bits(), (*it).image.bytesPerLine());
			return;
		}
	}

	QPixmap* pixmap = new QPixmap(width,height);
	QPainter* painter = new QPainter(pixmap);
	QMenuItem* mi = new QMenuItem();
	QMenuBar* mb = new QMenuBar(NULL);

	QStyle::SFlags sflags = QStyle::Style_Active | QStyle::Style_HasFocus;

	switch (state_type)
	{
	case GTK_STATE_ACTIVE: 		sflags |= QStyle::Style_Down;		break;
	case GTK_STATE_PRELIGHT:	sflags |= QStyle::Style_MouseOver;	break;
	case GTK_STATE_SELECTED:	sflags |= QStyle::Style_HasFocus;	break;
	case GTK_STATE_INSENSITIVE:	sflags = QStyle::Style_Default;		break;
	}

	qApp->style().drawControl(QStyle::CE_MenuBarItem, painter, mb, QRect(0,0,width,height), qApp->palette().active(), sflags, QStyleOption(mi));

	QImage image = pixmap->convertToImage().swapRGB();
	delete painter;
	delete pixmap;
	delete mi;
	delete mb;

	buffer.append(BufferedImage(BufferedImage::MenuItem, state_type, width, height, image));

	gdk_draw_rgb_32_image(window, style->bg_gc[state_type], x, y, width, height, GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
}

void drawMenuItem(GdkWindow * window, GtkStyle * style, GtkStateType state_type, int x, int y, int width, int height)
{
	gdk_draw_rectangle(window, style->bg_gc[state_type], true, x, y, width, height);
}

void setColors(GtkStyle* style)
{
	QWidget* meep = new QWidget(NULL);
	meep->polish();
	bool useBg = false;
	QImage image;

	if ((meep->paletteBackgroundPixmap()) && (!meep->paletteBackgroundPixmap()->isNull()))
	{
		meep->paletteBackgroundPixmap()->save("/home/david/meep.bmp","BMP");
		image = meep->paletteBackgroundPixmap()->convertToImage().swapRGB();
		useBg = true;
	}

	delete meep;

	for (uint i=0 ; i<5 ; i++)
	{
		switch (i)
		{
		case GTK_STATE_NORMAL:
		case GTK_STATE_INSENSITIVE:
			style->bg[i].red = qApp->palette().active().background().red()*257;
			style->bg[i].green = qApp->palette().active().background().green()*257;
			style->bg[i].blue = qApp->palette().active().background().blue()*257;
			break;
		case GTK_STATE_ACTIVE:
		case GTK_STATE_PRELIGHT:
		case GTK_STATE_SELECTED:
			style->bg[i].red = qApp->palette().active().highlight().red()*257;
			style->bg[i].green = qApp->palette().active().highlight().green()*257;
			style->bg[i].blue = qApp->palette().active().highlight().blue()*257;
			break;
		}

		style->fg[i].red = qApp->palette().active().foreground().red()*257;
		style->fg[i].green = qApp->palette().active().foreground().green()*257;
		style->fg[i].blue = qApp->palette().active().foreground().blue()*257;
		style->base[i].red = qApp->palette().active().base().red()*257;
		style->base[i].green = qApp->palette().active().base().green()*257;
		style->base[i].blue = qApp->palette().active().base().blue()*257;
		style->base[i].red = qApp->palette().active().base().red()*257;
		style->base[i].green = qApp->palette().active().base().green()*257;
		style->base[i].blue = qApp->palette().active().base().blue()*257;
		style->text[i].red = qApp->palette().active().text().red()*257;
		style->text[i].green = qApp->palette().active().text().green()*257;
		style->text[i].blue = qApp->palette().active().text().blue()*257;

		if (useBg)
		{
			GdkPixmap* pixmap = gdk_pixmap_new(NULL, image.width(), image.height(), style->depth);
			GdkGCValuesMask gc_values_mask = GDK_GC_FOREGROUND;
			GdkGCValues gc_values;
			gc_values.foreground = style->fg[0];
			GdkGC* gc = gtk_gc_get (style->depth, style->colormap, &gc_values, gc_values_mask);
			gdk_draw_rgb_32_image(pixmap, gc, 0, 0, image.width(), image.height(), GDK_RGB_DITHER_NONE, image.bits(), image.bytesPerLine());
			if (pixmap) style->bg_pixmap[i] = pixmap;
			gtk_gc_release(gc);
		}
	}
}
